# -*- mode: python; indent-tabs-mode: nil; tab-width: 4 -*-
# vim: set tabstop=4 shiftwidth=4 expandtab:

##############################################################################
# otr_staff.py
#
# toom, 2004-06-01
#
##############################################################################

from gamera.core import *
from gamera.plugins.listutilities import median
from gamera.plugins.image_utilities import union_images
#from gamera.toolkits.musicstaves import MusicStaves_rl_simple as MusicStaves
from gamera.toolkits.musicstaves import MusicStaves_linetracking as MusicStaves

class TabPage:
##############################################################################
    """Stores a tablature and its information (e.g. staff position).
    
An object of this class provides the following functionality:

- Staff line removal
- Return the information *staff number* and *course* for a given (x,y) position
- Return a list of staff bounding boxes
- Removal of regions outside any staff

.. note:: The removal of the staff lines is done by the inherited
          *MusicStaves* object. Thus, the *MusicStaves* toolkit is required.

A *TabPage* object has several public properties, that can be manipulated by
the user. The following table gives a detailed view on this.

+-----------------------+---------------------------------------------------+
| ``music_staves``      | The inherited *MusicStaves* object.               |
+-----------------------+---------------------------------------------------+
| ``image``             | A reference to ``music_staves.image``             |
|                       | (for convenience).                                |
+-----------------------+---------------------------------------------------+
| ``staffline_height``  | A reference to ``music_staves.staffline_height``  |
+-----------------------+---------------------------------------------------+
| ``staffspace_height`` | A reference to ``music_staves.staffspace_height`` |
+-----------------------+---------------------------------------------------+
| ``on_lines``          | Determines the position of the symbols            |
|                       | (see the constructor_ for details).               |
+-----------------------+---------------------------------------------------+
| ``tab_type``          | Determines the type of the tablature              |
|                       | (see the constructor_ for details).               |
+-----------------------+---------------------------------------------------+
| ``num_lines``         | The number of lines per staff.                    |
|                       | (only set after remove_staves_).                  |
+-----------------------+---------------------------------------------------+

.. _constructor: gamera.toolkits.otr.otr_staff.TabPage.html#init
.. _remove_staves: gamera.toolkits.otr.otr_staff.TabPage.html#remove-staves

:Author: Christoph Dalitz and Thomas Karsten
"""

    ######################################################################
    # constructor
    # 
    # img: Image
    # online:
    # tab_type:
    # staffline_height:
    # staffspace_height:
    #
    def __init__(self, img, online=0, tab_type='french', staffline_height=0, staffspace_height=0):
        """Constructs and returns a *TabPage* object.
        
Signature:

  ``__init__(image, online=0, tab_type='french', staffline_height=0, staffspace_height=0)``

with

  *image*:
    Onebit or greyscale image a *TabPage* should be created of.

  *online*:
    Determines the position of the glyphs in a staff. When *online* is set
    to 1 the tab symbols are supposed to be on the staff lines. A value of 0
    indicates that the tab symbols are located between the staff lines.

  *tab_type*:
    Classifies the type of the tablature. Possible values are ``french`` and
    ``italian``.

  *staffline_height*:
    Vertical thickness of a staff line. When not specified here, it will be
    detected automatically.

  *staffspace_height*:
    Thickness of the vertical space between adjacent staff lines. When not
    specified here, it will be detected automatically.
"""

        self.image=None      # image of the tablature (reference to
                     # self.music_staves.image, see below)
        self.stafflist=[]    # list of staffobj (staffno, yposlist)
        self.staffranges=[]  # list of [start, end] of staves
                     # y-dimensions incl. flags

        self.on_lines=online # tab symbols in courses

        if tab_type not in ['french', 'italian']:
            raise RuntimeError, tab_type + ": Unknown Tabtype"
        else:
            self.tabtype=tab_type

        self.num_lines = 0 # will be set in remove_staves

        # use external class module for removing the staves
        self.music_staves=MusicStaves(img, staffline_height,\
                staffspace_height)

        # references to important information
        self.image=self.music_staves.image
        self.staffline_height=self.music_staves.staffline_height
        self.staffspace_height=self.music_staves.staffspace_height

    #####################################################################
    # IntVector get_staff_range(staff, staffspace_height, staffline_height)
    #
    # parameters:
    #      staff: int vector of staffline y-positions
    #      staffspace_height: height of the space between the stafflines
    #      staffline_height: height of a staffline
    #
    # returns an intvector containing the range of staff. the first value
    # contains the y-position of the start, the second value contains the
    # y-position of the end of staff.
    #
    # toom, 2004-07-01
    #
    def __get_staff_range(self, staff, staffspace_height, staffline_height):
        n_stafflines=len(staff)
        start=staff[0]-3*staffspace_height

        if start < 0:
            start=0

        end=staff[0]+n_stafflines*(staffspace_height+staffline_height)

        if end > self.image.height:
            end=self.image.height

        return [start, end]

    ######################################################################
    # get_staffboxes(self)
    #
    # possible values for tightness:
    #      0 = only the region within the stafflines (tight staff)
    #      1 = wider border (including flags etc) around staff
    #
    def get_staffboxes(self, tightness=1):
        """Returns a list of rectangular bounding boxes, that contains a bounding box for each staff.

Signature:

  ``get_staffboxes(tightness=1)``

with

  *tightness*:
    Controls the y-borderwidth around the staff. The output of this function
    for the two possible values 0 and 1 can be seen in the image below.

If there are *n* tablature systems on the page, a sequence of *n* objects
with the properties ``x``, ``y``, ``nrows`` and ``ncols`` are returned. These
properties describe the tablature system rectangular bounding box (dependent
of the value of *tightness*) as shown in the figure below.

.. image:: images/staffboxes.png

"""
        class staffbox:
            x=0; y=0; ncols=0; nrows=0
        staffboxes=[]
        # TODO: x-positionen setzen
        #       Idee: finde extreme CC's die Staffs enthalten
        if (tightness == 1):
            for s in self.staffranges:
                sb = staffbox()
                sb.x = 0
                sb.y = s[0]
                sb.ncols = self.image.ncols
                sb.nrows = s[1] - s[0]
                staffboxes.append(sb)
        else:
            for s in self.stafflist:
                sb = staffbox()
                sb.x = 0
                sb.y = s.yposlist[0]
                sb.ncols = self.image.ncols
                sb.nrows = s.yposlist[-1] - s.yposlist[0]
                staffboxes.append(sb)

        return staffboxes


    ######################################################################
    # remove_staves
    #
    # Christoph Dalitz
    #
    def remove_staves(self, crossing_symbols='auto', num_lines=0):
        """Detects and removes staff lines from a music/tablature image.

Signature:

  ``remove_staves(crossing_symbols='auto', num_lines=0)``

with

  *crossing_symbols*:
    Determines which symbols crossing staff lines the removal should try
    to keep intact. In addition to the possible values are ``all``, ``bars``
    and ``none`` (see the MusicStaves module for details), the value ``auto``
    is supported for a setting based upon onlines.
  
  *num_lines*:
    Number of lines within one staff. A value of zero means, that this number
    should be guessed automatically.
"""
        if crossing_symbols=='auto':
            if self.on_lines==1:
                crossing_symbols='all'
            else:
                crossing_symbols='bars'

        self.music_staves.remove_staves(\
                crossing_symbols=crossing_symbols,\
                num_lines=num_lines)
        self.stafflist=self.music_staves.get_staffpos()
        if num_lines > 0:
            self.num_lines = num_lines
        else:
            self.num_lines = median([len(s.yposlist) for s in self.stafflist])

        # for interrupted systems, we can have more than one stafflist per line
        # if this occurs, only keep the y-positions of the widest
        newstafflist = [self.stafflist[0]]
        staffno = 1
        for i, s in enumerate(self.stafflist):
            if i == 0:
                continue
            if newstafflist[-1].staffrect.intersects_y(s.staffrect):
                r1 = s.staffrect
                r2 = newstafflist[-1].staffrect
                if r1.width > r2.width:
                    newstafflist[-1] = s
                newstafflist[-1].staffrect.ul_x = min([r1.ul_x,r2.ul_x])
                newstafflist[-1].staffrect.lr_x = max([r1.lr_x,r2.lr_x])
                newstafflist[-1].staffno = staffno
            else:
                staffno += 1
                newstafflist.append(s)
                newstafflist[-1].staffno = staffno
        self.stafflist = newstafflist
            

        # compute staff ranges (includes flags above staff)
        self.staffranges = []
        for s in self.stafflist:
            srange = self.__get_staff_range(s.yposlist,\
                         self.music_staves.\
                         staffspace_height,\
                         self.music_staves.\
                         staffline_height)
            self.staffranges.append(srange)

        # print some information on found staves
        l=len(self.stafflist)
        if l != 1:
            print "Found", l, "staves:"
        else:
            print "Found", l, "staff:"

        for i,s in enumerate(self.stafflist):
            print "    Staff", s.staffno, ":", s.yposlist,\
                  "   Range:", self.staffranges[i]

        # make self.image a reference to the actual
        # image of the MusicStaves object
        self.image=self.music_staves.image


    ######################################################################
    #
    def remove_outliers(self):
        """TODO: Removes regions outside of tablature staves like
texts, page numbers etc.
"""
        pass

    ######################################################################
    # get_staffcourse(x, y)
    #
    # returns 2 values [int staff, int course] for a symbol at
    # position (x, y). staff and course are counted from top to bottom,
    # starting with 1 each. symbols that are above the first course
    # (e. g. flags) get the course number 0.
    # an empty array [] is returned, when (x, y) is out of range.
    #
    # toom, 2004-06-01
    #
    def get_staffcourse(self, x, y):
        """Returns the two values [Int *staff*, Int *course*] for a
symbol at position (*x*, *y*). Both staff and course are counted from top to
bottom, starting with 1 each. Symbols that are above the first course (e.g.
flags) get the course number 0. An empty array [] is returned, when (*x*, *y*)
is out of range.

Signature:

  ``get_staffcourse(x, y)``

with

  (*x*, *y*):
    x- and y-position of the symbol to query.
"""

        if self.on_lines:
            y-=self.music_staves.staffspace_height/2

        n_staves = len(self.stafflist)
        
        #
        # the method 'remove_staves' should already have been applied 
        #
        if n_staves == 0:
            return []

        ret_value=[]
        i=0

        # glyph is above the first staff
        if y < self.staffranges[0][0]:
            return []

        while i < n_staves:

            # get the range of the current staff
            [start, end] = self.staffranges[i]
    
            # is the glyph located in this very staff?
            if y >= start and y <= end:
    
                ret_value.append(i+1)

                # now: find the *exact* staffline
                virtual_first_staffline=\
                        self.stafflist[i].yposlist[0]\
                        -self.music_staves.\
                        staffspace_height

                if y < virtual_first_staffline:
                    if self.tabtype == "french":
                        ret_value.append(0)
                        return ret_value
                    elif self.tabtype == "italian":
                        ret_value.append(self.num_lines+1)
                        return ret_value

                if self.tabtype == "french":
                    staffLineNo=1
                elif self.tabtype == "italian":
                    staffLineNo=self.num_lines

                for line in self.stafflist[i].yposlist:
                    if y < line:
                        ret_value.append(staffLineNo)
                        return ret_value
                    else:
                        if self.tabtype == "french":
                            staffLineNo+=1
                        elif self.tabtype == "italian":
                            staffLineNo-=1

                # do not forget glyphs below the last
                # staffline
                ret_value.append(staffLineNo)
                return ret_value

            # check the next staff
            i+=1

        # (x, y) is out of range
        #print "out of range! y: ", y, "start", start, "end", end
        return []

    ######################################################################
    # remove_barlines
    #
    # Christoph Dalitz
    #
    def remove_barlines(self):
        """Removes bar lines from a music/tablature image and returns them as 
a list of Cc's. the returned Cc's are classified heuristically as \"bar\".

Signature:

  ``remove_barlines()``

Bar line candidates are built from filaments of short horizontal runlength
by a number of heuristic rules. As these rules make use of the staffline
positions, this method can only be called after *remove_staves*.

As the bar lines are removed from ``tab.image``, the returned Cc's point
to a different image (Python's garbage collector will take care of keeping or
deleting this image as necessary).
"""
        #
        # step 1: find barlines inside stafflines
        #
        bar_width = self.staffline_height
        img = self.image.image_copy()
        # remove wide runs
        img.filter_wide_runs(int(2.5*bar_width),"black")
        ccs = img.cc_analysis()
        newccs = []
        # remove all non tall Ccs
        for cc in ccs:
            if (cc.height>2) and float(cc.nrows)/cc.ncols>=2.5:
                newccs.append(cc)
        ccs = newccs
        newccs = []
        # keep only fragments that intersect the stafflines
        staffs = self.get_staffboxes(tightness=0)
        staffrects = [Rect((s.x,s.y),Dim(s.ncols,s.nrows)) for s in staffs]
        for cc in ccs:
            intersects = False
            for i,sr in enumerate(staffrects):
                if cc.intersects(sr):
                    cc.staffno = i
                    intersects = True
                    break
            if intersects:
                newccs.append(cc)
        ccs = newccs
        # group bar candidates to chord groups
        staffno = 0
        thisstaff = [c for c in ccs if c.staffno == staffno]
        chords = []
        while len(thisstaff) > 0:
            thisstaff.sort(lambda s1, s2: s1.center_x-s2.center_x)
            for i,cc in enumerate(thisstaff):
                if i == 0:
                    chord = [cc]
                    continue
                intersects = False
                for c in chord:
                    if cc.intersects_x(c):
                        intersects = True
                        break
                if not intersects:
                    chords.append(chord)
                    chord = [cc]
                else:
                    chord.append(cc)
            chords.append(chord)
            staffno += 1
            thisstaff = [c for c in ccs if c.staffno == staffno]
        # heuristic criteria which groups are bars
        barlines = []
        for c in chords:
            maxheight = 0
            top = img.nrows
            bot = 0
            for cc in c:
                if cc.nrows > maxheight:
                    maxheight = cc.nrows
                if cc.ul_y < top:
                    top = cc.ul_y
                if cc.lr_y > bot:
                    bot = cc.lr_y
            # at least one part must be higher than t1 * staffspace_height
            # total hight must be higher than t2 * staffspace_height
            if maxheight > 1.2*self.staffspace_height and bot-top > 2.5*self.staffspace_height:
                    mlcc = MlCc(c)
                    cc = mlcc.convert_to_cc()
                    cc.classify_heuristic("bar")
                    cc.staffno = c[0].staffno
                    barlines.append(cc)

        #
        # step 2: add overlapping bar line parts outside staff
        #
        extendedstaffrects = [Rect((s.ul_x,max([s.ul_y-self.staffspace_height,0])),(s.lr_x,s.lr_y+self.staffspace_height)) for s in staffrects]
        ccs = self.image.cc_analysis()
        ccs = [cc for cc in ccs if cc.ncols < int(2.5*bar_width)]
        newccs = []
        for cc in ccs:
            intersects = False
            for i,sr in enumerate(staffrects):
                if cc.intersects(sr):
                    cc.staffno = i
                    intersects = True
                    break
            if not intersects:
                # assign to closest staff
                for i,sr in enumerate(extendedstaffrects):
                    if cc.intersects(sr):
                        cc.staffno = i
                        newccs.append(cc)
                        break
        ccs = newccs
        # merge outlier candidates with overlapping barlines
        staffno = 0
        allccs = ccs + barlines
        thisstaff = [c for c in allccs if c.staffno == staffno]
        chords = []
        while len(thisstaff) > 0:
            thisstaff.sort(lambda s1, s2: s1.center_x-s2.center_x)
            for i,cc in enumerate(thisstaff):
                if i == 0:
                    chord = [cc]
                    continue
                intersects = False
                for c in chord:
                    if cc.intersects_x(c):
                        intersects = True
                        break
                if not intersects:
                    chords.append(chord)
                    chord = [cc]
                else:
                    chord.append(cc)
            chords.append(chord)
            staffno += 1
            thisstaff = [c for c in allccs if c.staffno == staffno]
        barlines = []
        for c in chords:
            if len(c) == 1:
                if c[0].match_id_name("bar"):
                    barlines.append(c[0])
            else:
                isbar = False
                for g in c:
                    if g.match_id_name("bar"):
                        isbar = True
                        break
                if isbar:
                    cc = union_images(c)
                    cc.classify_heuristic("bar")
                    barlines.append(cc)
        # remove barlines in original image
        for cc in barlines:
            self.image.highlight(cc,0)
        return barlines

