from gamera.plugin import *
import _otr_staff_plugins

class TabPage(PluginFunction):
	"""Creates a TabPage__ object.

.. __: gamera.toolkits.otr.otr_staff.TabPage.html#tabpage
"""
	pure_python = 1
	self_type = ImageType([ONEBIT, GREYSCALE])
	args = Args([Int('online', default=0),\
			String('tab_type'),\
			Int('staffline_height', default=0),\
			Int('staffspace_height', default=0)])
	return_type = Class("TabPage")

	def __call__(image, online=0, tab_type='french', line_h=0, space_h=0):
		from gamera.toolkits.otr import otr_staff
		return otr_staff.TabPage(image, online, tab_type, line_h,\
				space_h)

	__call__ = staticmethod(__call__)

class deskew(PluginFunction):
	"""Detects a rotation angle and corrects it with the aid of
skewed projections.

When the image is large, it is temporarily scaled down to a staffline
height of two pixels for performance reasons.
"""
	pure_python = 1
	category="OTR/preprocessing"
	self_type = ImageType([ONEBIT, GREYSCALE])
	return_type = ImageType([ONEBIT, GREYSCALE])
	author = "Christoph Dalitz"

	def __call__(self):

		if self.data.pixel_type != ONEBIT:
			bwtmp = self.to_onebit()
			rotatebordervalue = 255
		else:
			bwtmp = self
			rotatebordervalue = 0
		# rescale large images for performance reasons
		slh = bwtmp.most_frequent_run('black','vertical')
		if (slh > 3):
			print "scale image down by factor %4.2f" % (2.0/slh)
			bwtmp = bwtmp.scale(2.0/slh, 0)
		# find and remove rotation
		skew = bwtmp.rotation_angle_projections(-2,2)
		if (abs(skew[0]) > abs(skew[1])):
			print ("rot %0.4f degrees found ..." % skew[0]),
			retval = self.rotate(skew[0], rotatebordervalue)
			print "and corrected"
		else:
			print "no rotation found"
			retval = self.image_copy()
		return retval

	__call__ = staticmethod(__call__)


class smoothen(PluginFunction):
	"""Convolves the image with a smoothing kernel, which is adapted to
the characteristic image dimensions: it is *n/2+1* pixels high and *n* pixels
wide, where *n* is the staffline height of the image. This shape is chosen
in such a way that stafflines become better connected.
"""
	pure_python = 1
	category="OTR/preprocessing"
	self_type = ImageType([GREYSCALE, GREY16])
	return_type = ImageType([GREYSCALE, GREY16])
	author = "Christoph Dalitz"

	def __call__(self):

		def normalize_kernel(kernel):
			normfactor = 0
			for row in kernel:
				for c in row:
					normfactor += c
			return [[1.0*c/normfactor for c in row] for row in kernel]
	
		#print "Smoothen..."
	
		img = self.image_copy()

		# get characteristic page dimensions
		bw0 = img.to_onebit()
		slh = bw0.most_frequent_run('black','vertical')
	
		# build convolution kernel and apply
		n = (slh / 2) * 2 + 1
		if n > 3:
			#kernel = [[0.5] + [1.0]*(n-2) + [0.5]] * (n/2 + 1)
			m = n/5
			kernel1 = [[0.5]*m + [1.0]*(n-2*m) + [0.5]*m] * (n/2 + 1)
		else:
			kernel1 = [[1.0] * n] * (n/2 + 1)
		# alternative more coarse kernel
		kernel2 = [[1.0] * n] * n
		kernel = normalize_kernel(kernel1)
		imgnew = img.convolve(kernel)
	
		return imgnew

	__call__ = staticmethod(__call__)


class remove_pictures(PluginFunction):
	"""Removes large CC's that do not \"look like\" staff sections. This
decision is based upon staffspace_height (most frequent white vertical run):

- first the global staffspace height (*ssh*) is computed

- then all CC's are removed which are larger than *height_factor* \* *ssh*
  and whose most frequent white vertical run is different from *ssh*
"""
	pure_python = 1
	category="OTR/preprocessing"
	self_type = ImageType([ONEBIT])
	args = Args([Int("height_factor", default=4)])
	return_type = ImageType([ONEBIT])
	author = "Christoph Dalitz"

	def __call__(self, height_factor=4):

		from gamera.core import Point
		img=self.image_copy()

		# get characteristic staff feature
		sph = img.most_frequent_run('white','vertical')
		ccs = img.cc_analysis()
		large_ccs = [c for c in ccs\
				if ((c.nrows > height_factor*sph) and\
				(c.ncols > height_factor*sph))]

		for c in large_ccs:
			if abs(c.most_frequent_run('white','vertical') - sph) > 2:
				img.draw_filled_rect(\
					Point(c.offset_x,c.offset_y),\
					Point(c.offset_x+c.ncols,c.offset_y+c.nrows),\
					0)
		img.otr_reset_onebit_image()
		return img

	__call__ = staticmethod(__call__)


class otr_reset_onebit_image(PluginFunction):
	"""Resets all black pixel values in a onebit image to one.
This can be necessary e.g. after a CC analysis which sets black
pixels to some other label value.
"""
	category="OTR/preprocessing"
	self_type = ImageType([ONEBIT])
	author = "Christoph Dalitz"





#----------------------------------------------------------------
class smooth_onebit(PluginFunction):
    """Smooth the image by first despeckling with *speckle_size* and then
filtering out white speckles. The despeckling can be skipped by giving a
zero speckle size. The filtering method depends on the
option *filter*:

  0: applies a median filter (*rank(5)*)

  1: closes the image with a 3x3 black structuring element.

  2: like despeckle, but with white speckles
"""
    category = "OTR/preprocessing"
    pure_python = 1
    self_type = ImageType([ONEBIT])
    args = Args([Int('speckle_size', default=3),
                 Choice("filter", ["median", "closing", "wspeckle"])])
    return_type = ImageType([ONEBIT])

    def __call__(self,speckle_size=3,filter=0):
        from gamera.core import Point, Image
        image=self.image_copy()
        if speckle_size > 0:
            image.despeckle(speckle_size)
        if filter == 0:
            filtered = image.rank(5)
        elif filter == 1:
            origin = Point(1,1)
            se = Image(Point(0,0),Point(3,3),ONEBIT)
            se.fill(1)
            filtered = image.dilate_with_structure(se,origin)
            filtered = filtered.erode_with_structure(se,origin)
        else:
            image.invert()
            image.despeckle(5)
            image.invert()
            filtered = image
            
        return filtered
    __call__ = staticmethod(__call__)


#-----------------------------------------------------------------
class remove_copy_border(PluginFunction):
    """Removes large CCs that are possibly black borders from scanning.

Depending on *feature*, large CCs are identified as follows:

 *size*:
   CCs with a height greater than the 20 times the median height or a width
   greater than 10 times the median width of all CCs with width/height > 3
   (this median is an estimator for the *oligon_width*) are removed

 *black_area*:
   CCs with a black_area greater than 20 times the median black area
   are removed

"""
    category = "OTR/preprocessing"
    pure_python=1
    self_type = ImageType([ONEBIT,GREYSCALE,RGB])
    return_type=ImageType([ONEBIT])
    args = Args([ChoiceString('feature', ['size','black_area'], default='size')])
    author="Christoph Dalitz"

    def __call__(self, feature='size'):
        onebit=self.to_onebit()
        ccs=onebit.cc_analysis()
        if feature == 'size':
            widths = [c.ncols for c in ccs if c.ncols/c.nrows > 3]
            widths.sort()
            max_width = 10 * widths[len(widths)/2]
            heights = [c.nrows for c in ccs]
            heights.sort()
            max_height = 20 * heights[len(heights)/2]
            for c in ccs:
                if c.ncols > max_width or c.nrows > max_height:
                    c.fill_white()
        else:
            areas = []
            for c in ccs:
                # store the value with the cc for performance reason
                c.black_ar = c.black_area()[0]
                areas.append(c.black_ar)
            areas.sort()
            max_area = 20 * areas[len(areas)/2]
            for c in ccs:
                if c.black_ar > max_area:
                    c.fill_white()

        return onebit

    __call__ = staticmethod(__call__)



#-------------------------------------------------------------------------
class preprocess(PluginFunction):
    """This is only a wrapper function for the other preprocessing plugins.
It provides a convenient GUI to select preprocessing operations
and run them all at once.
"""
    category = "OTR/preprocessing"
    pure_python = 1
    self_type = ImageType([ONEBIT])
    args = Args([Int('speckle_size', default=3),
                 Choice("smoothing_method", ["median", "closing", "wspeckle", "none"]),
                 Check('correct_rotation',default=False),
                 Check('remove_copy_border',default=False)])
    return_type= ImageType([ONEBIT])
    author = "Christine Pranzas"

    def __call__(self,speckle_size=3,smooth=0,correct_rotation=False,remove_copy_border=False):
        image=self.image_copy()
        if correct_rotation==True:
            image=image.correct_rotation()
        if smooth < 3:
            image=image.smooth_onebit(speckle_size,smooth)
        elif speckle_size > 0:
            image.despeckle(speckle_size)
        if remove_copy_border==True:
            image=image.remove_copy_border(feature)
        return image
        
    __call__ = staticmethod(__call__)



class TabPageModule(PluginModule):
	category = "OTR"
	cpp_headers=["otr_staff_plugins.hpp"]
	functions = [TabPage, deskew, smoothen,preprocess,\
		     smooth_onebit,remove_copy_border,\
		     remove_pictures, otr_reset_onebit_image]
	author = "Christoph Dalitz, Thomas Karsten"

module = TabPageModule()
