##############################################################################
# to_abcmusic.py
# Christoph Dalitz, 2006
# Christine Pranzas, 2007
# Christian Brandt, 2012
#
##############################################################################

## german_page
## is a page for german lute tabulature
from gamera import toolkit
import sys
from german_page import *
import plugins
from gamera.core import *
from gamera.args import *
from gamera import toolkit
from gamera.toolkits.otr.plugins import *
from gamera.core import *


debugout = []

def debug_print_glyphs(glyphs, prefix="DBG"):
    if prefix in debugout: # ["after singleglyph","lastx==g.x"]: #"2"]:
        print '\033[1;32m[DBG Glyphs %s]\033[1;m' % prefix
        for g in glyphs:
            print "     ", prefix, g.id_name, g.x_number, g.gridnr, g

def debug_print(a, prefix="DBG"):
    if prefix in debugout:
        print '\033[1;31m[DBG %s]\033[1;m' % prefix, a

class ToAbcmusic:
##############################################################################
    """Generates abc music code from tablature glyphs that are already
semantically grouped.

The code is generated in the constructor and stored in the property
*abc_music*.
"""

    ######################################################################
    # constructor
    #
    def __init__(self, glyphs=None,tabtype="french",tuning="gdAFCG,", key='auto', factor=0):
        """Gives an output of a classified an semantically interpreted
glyphlist as an *abc music*-formatted string.

Signature:

  ``init(glyphs, tabtype, tuning=\"gdAFCG,\", key=\"auto\", factor=0)``

with

  *glyphs*:
     classified and semantically interpreted glyphs. Depending on *tabtype*,
     this must be either the sorted list returned by
     TabGlypList.set_glyph_properties_ or the property *page*
     of a GermanGlyph_ object.

.. _TabGlypList.set_glyph_properties: gamera.toolkits.otr.otr_glyph.TabGlyphList.html#set-glyph-properties
.. _GermanGlyph: gamera.toolkits.otr.german_glyph.GermanGlyph.html

  *tuning*:
     The tuning of each course, from top to bottom. Each note is specified
     as *abc* code (sharp = prefix '^', flat = prefix '_', octave up = postfix
     single quote, octave down = postfix comma)

  *key*:
     Key signature, e.g. the value ``Eb`` stands for *E flat Major* or ``Cm``
     for *C Minor*. When set to ``auto`` the key is guessed automatically.

  *factor*:
     Logarithmic length factor for slowing down. For example, when set to
     ``2`` the sixteenth notes are printed as quarter notes.

The resulting abc string is stored in the public property *abc_music* and
can be given as input to the external program `abctab2ps`__.

.. __: http://www.lautengesellschaft.de/cdmm/

"""

        self.abc_music=[]
        self.glyphs=glyphs
        self.tabtype=tabtype
        self.tuning=tuning
        self.key=key
        self.factor=factor
        self.lastflagstr = ""

        if tabtype=="french" or tabtype=="italian":
            self.abc_music= self.to_abcmusic()

        elif tabtype=="german":
            self.abc_music= self.to_abcmusic_german()


    # some settings for tab -> music transcription
    #-------------------------------------------------------------------------------

    # map key -> accidentals
    keyaccs = {'C':[],'Am':[],\
           'F':['_b'],'Dm':['_b'],\
           'Bb':['_b','_e'],'Gm':['_b','_e'],\
           'Eb':['_b','_e','_a'],'Cm':['_b','_e','_a'],\
           'Ab':['_b','_e','_a','_d'],'Fm':['_b','_e','_a','_d'],\
           'G:':['^f'],'Em:':['^f'],\
           'D:':['^f','^c'],'Bm:':['^f','^c'],\
           'A:':['^f','^c','^g'],'F#m:':['^f','^c','^g'],\
           'E:':['^f','^c','^g','^d'],'C#m:':['^f','^c','^g','^d']}
    # map key -> scale
    keyscales = {'C':['c','^c','d','^d','e','f','^f','g','^g','a','_b','b'],\
             'Am':['c','^c','d','^d','e','f','^f','g','^g','a','_b','b'],\
             'F':['c','^c','d','_e','e','f','^f','g','^g','a','_b','b'],\
             'Dm':['c','^c','d','_e','e','f','^f','g','^g','a','_b','b'],\
             'Bb':['c','^c','d','_e','e','f','^f','g','_a','a','_b','b'],\
             'Gm':['c','^c','d','_e','e','f','^f','g','_a','a','_b','b'],\
             'Eb':['c','_d','d','_e','e','f','^f','g','_a','a','_b','b'],\
             'Cm':['c','_d','d','_e','e','f','^f','g','_a','a','_b','b'],\
             'Ab':['c','_d','d','_e','e','f','_g','g','_a','a','_b','b'],\
             'Fm':['c','_d','d','_e','e','f','_g','g','_a','a','_b','b'],\
             'G':['c','^c','d','^d','e','f','^f','g','^g','a','^a','b'],\
             'Em':['c','^c','d','^d','e','f','^f','g','^g','a','^a','b'],\
             'D':['c','^c','d','^d','e','f','^f','g','^g','a','^a','b'],\
             'Bm':['c','^c','d','^d','e','f','^f','g','^g','a','^a','b'],\
             'A':['c','^c','d','^d','e','^e','^f','g','^g','a','^a','b'],\
             'F#m':['c','^c','d','^d','e','^e','^f','g','^g','a','^a','b'],\
             'E':['^b','^c','d','^d','e','^e','^f','g','^g','a','^a','b'],\
             'C#m':['^b','^c','d','^d','e','^e','^f','g','^g','a','^a','b']}

    # scale index of notes
    # double notes mean "halftone between these notes"
    scalenote = ['c','cd','d','de','e','f','fg','g','ga','a','ab','b']
    # halftone array for tuning interpretation
    scalenum = {'c':0,'^c':1,'_d':1,'d':2,'^d':3,'_e':3,\
            'e':4,'f':5,'^f':6,'_g':6,'g':7,'^g':8,\
            '_a':8,'a':9,'^a':10,'_b':10,'b':11}
            # fret symbols
    fretnum = {'a':0,'b':1,'c':2,'d':3,'e':4,\
               'f':5,'g':6,'h':7,'i':8,'k':9,\
               'l':10,'m':11,'n':12,'o':13,'p':14}
    song_header = "X:1\n"+\
           "%\n"+\
           "%%MIDI program 25\n"+\
           "%%titleleft true\n"+\
           "%\n"+\
           "T:OTR result\n"+\
           "C:autmatic music transcription\n"

    class tuningcourse:
        def __init__(self, note, octave):
            self.note=note; self.octave=octave

    def build_chord(self,chord, flags,div):
            flag_str = self.lastflagstr

            flag = []
#            print len(flags)
#            print "DBG1", flags
            if len(flags) > 0:
                flag = flags.pop(0)
            if len(flag) > 0:
                flag_str = ""

                ff = flag[1]
                if ((flag[1] % 3) == 0):
                    key = str(flag[1])
                    flag_str="3/"+str((dotflags[key]/div))

                elif flag[0] == 1:
                    if flag[1] > div:
                        while (ff > div):
                            flag_str+='/'
                            ff/=2
                    else:
                        flag_str=str(int(div/flag[1]))
                else:
                    ff/=div
                    flag_str=str(flag[0])+"/"+str(ff)
 #               print "flagstr", flag_str
            self.lastflagstr = flag_str
            if len(flags) == 0:
                #               print "withspace"
                return "[" + chord[0] + flag_str + \
                   "".join(chord[1:]) + "] "

            return "[" + chord[0] + flag_str + \
                   "".join(chord[1:]) + "]"




    # for remembering set accidentals in current bar

    def lastsetacc(self,note,setaccs):
        sl = len(setaccs)
        for i in range(0,sl):
            sa = setaccs[sl-i-1][1]
            if sa == note:
                return sa[0]
        return ""

    def accidentalize(self,note,setaccs):
        accs = self.keyaccs[self.key]

        # compute number of sharps (negative for b's)
        if len(accs) == 0:
            nsharps = 0
        else:
            nsharps = len(accs)
            if accs[0][0] == "_":
                nsharps = nsharps * (-1)

        # check natural notes
        if len(note) == 1:
            sa = self.lastsetacc(note,setaccs)
            if len(sa) > 0:
                if sa == "=":
                    return note
                else:
                    return ("=" + note)
            for a in accs:
                if note == a[1]:
                    setaccs.append("=" + note)
                    return ("=" + note)
            return note

        # check chromatic notes
        sa0 = self.lastsetacc(note[0],setaccs)
        sa1 = self.lastsetacc(note[1],setaccs)
        for a in accs:
            if a[0]=='^' and a[1]==note[0]:
                if sa0 != '_' and sa0 != '=':
                    return note[0]
                else:
                    return a
            elif a[0]=='_' and a[1]==note[1]:
                if sa0 != '^' and sa0 != '=':
                    return note[1]
                else:
                    return a
        # other accs are resolved according key scale
        ks = self.keyscales[self.key]
        return (ks[self.scalenote.index(note)])

    #----------------------------------------------

    def test_b_sharp(self,notes):
        # test b's
        nb = 0
        for b in ['b','e','a','d']:
            natural = 0; flat = 0
            for n in notes:
                if len(n) > 1:
                    if n[1] == b:
                        flat += 1
                elif n == b:
                    natural += 1
            if flat == 0 or natural > flat:
                break
            else:
                nb += 1
        if nb > 0:
            return ['F','Bb','Eb','Ab'][nb-1]

        # test sharp's
        ns = 0
        for s in ['b','e','a','d']:
            natural = 0; sharp = 0
            for n in notes:
                if len(n) > 1:
                    if n[0] == s:
                        sharp += 1
                elif n == b:
                    natural += 1
            if sharp == 0 or natural > sharp:
                break
            else:
                ns += 1
        if ns > 0:
            return ['G','D','A','E'][nb-1]

        # no accidentals found:
        return 'C'

    #-------------------------------------------------------

    def get_tuninglist(self):
        # parse tuning into list of tuningcourses
        tuninglist=[]
        note=""
        notefound=0
        octave=1
        for c in self.tuning:
            if c.isalpha() or c=='^' or c=='_':
                if notefound:
                    tc = self.tuningcourse(note,octave)
                    tuninglist.append(tc)
                    note=""; octave=1; notefound=0
                note += c.lower()
                if c.isalpha():
                    notefound = 1
                    if c.isupper():
                        octave -= 1
            elif c==",":
                octave -= 1
            elif c=="'":
                octave += 1
            else:
                print "invalid char in tuning: \"%c\"" % c
        if notefound:
            tc = self.tuningcourse(note,octave)
            tuninglist.append(tc)

        return tuninglist


#---------------------------------------------------------------------------
    def to_abcmusic(self):
        def transcribe(fret, course, tuning):

            if course > len(tuning):
                return ""

            # compute note and octave
            octave = tuning[course-1].octave
            note = self.scalenum[tuning[course-1].note]
            octave += (note + self.fretnum[fret]) / 12
            note = (note + self.fretnum[fret]) % 12

            musicnote = self.scalenote[note]
            #print "note '" + musicnote + "'",
            musicnote = self.accidentalize(musicnote,setaccs)
            #print "-> '" + musicnote + "'"

            if octave <= 0:
                musicnote = musicnote.upper()
                for i in range(octave,0):
                    musicnote += ","
            elif octave > 1:
                for i in range(1,octave):
                    musicnote += "'"

            return musicnote

        # guess the key such that minimal accidentals will result
        def guesskey(tuning):

            # build up list of all notes (without octave info)
            notes = []
            for g in self.glyphs:
                if g.match_id_name("*fret*") and\
                        g.course <= len(tuninglist):
                    f=g.id_name[0][1].split('.')
                    fretletter=f[-1]
                    noteind = (self.scalenum[tuning[g.course-1].note] \
                           + self.fretnum[fretletter]) % 12
                    notes.append(self.scalenote[noteind])

            # test b's and sharp's
            return self.test_b_sharp(notes)

        def get_flagcourse():
            #gets major course containing flags and creates a map of flags sorted by staff
            courses = {}
            flagmap = {}
            lastflagidx=-1
            for g in self.glyphs:
                if g.match_id_name("*flag*"):
                    lastflagidx += 1
                    if g.staff >= 0:
                        if courses.has_key(g.course):
                            courses[g.course]+=1
                        else:
                            courses[g.course] = 1

                        if not flagmap.has_key(g.staff):
                            flagmap[g.staff] = [g]
                        else:
                            flagmap[g.staff].append(g)

            count = 0
            course = -1
            for c in courses:
                if courses[c] > count:
                    count = courses[c]
                    course = c

            return course,flagmap

         # parse tuning into list of tuningcourses
        tuninglist=self.get_tuninglist()


        # guess key if wished
        if self.key == 'auto':
            self.key = guesskey(tuninglist)
            print "key guessed as '" + self.key + "'"

        chord=[]
        course=1
        cur_staff=1
        cur_x_number=1
        flags=[]
        self.lastflagstr = ""
        dot=""
        div=4
        Lfieldval = 1.0/div * (2**self.factor)
        setaccs = []
        # write abc header
        song = self.song_header

        if Lfieldval >= 1:
            song += "L:" + str(int(Lfieldval)) + "/1\n"
            song += "Q:" + str(int(Lfieldval)) + "/1=50\n"
        else:
            song += "L:1/" + str(int(1.0/Lfieldval)) + "\n"
            song += "Q:1/" + str(int(1.0/Lfieldval)) + "=50\n"
        song += "K:" + self.key + "\n"+\
            "%\n"

        flagcourse,flags_by_staff = get_flagcourse()
        debug_print("flagcourse is %d" % flagcourse, "flagcourse")

        for i, g in enumerate(self.glyphs):

            # ignore characters outside of the staves
            if g.staff < 0 or g.course < 0:
                continue

            # start a new staff
            if g.staff != cur_staff:
                #self.lastflagstr = ""
                if len(chord) > 0:
                    song+=self.build_chord(chord, flags, div)
                    chord=[]

                song+="\n%\n"
                cur_staff=g.staff

            if g.x_number != cur_x_number:
                if len(chord) > 0:
                    song+=self.build_chord(chord, flags, div)
                    chord=[]

                course=1
                dot=""
                #flag=[] (must not be reset, but remembered)
                cur_x_number=g.x_number

            #
            # test for recognized glyphs
            #
            if g.match_id_name("bar"):
                song+="|"
                setaccs = []
                flag_course=-1

                # produce nice formatted output
                if i < len(self.glyphs)-1:
                    if self.glyphs[i+1].staff == cur_staff:
                        song+="\\\n\t"
            elif g.match_id_name("*flag.*.*"):
                f = g.id_name[0][1].split('.')[1:]

                #remove incomplete flags from last runs
                if len(flags) > 0:
                    # song += "[y]"*len(flags) + " "
                    flags = []
                i = 0
                while i < len(f):
                    debug_print((i, f[i]), "to_abcmusic")
                    #        print i,f[i], f[i+1]
                    if f[i].startswith("dot"):
                        dotidx = int(f[i][3:])
                        if dotidx < len(flags):
                            #denom = flags[-dotidx][0]
                            #flags[-dotidx][0] = flags[-dotidx][0]*2+denom
                            flags[dotidx-1][0] *= 3
                            flags[dotidx-1][1] *= 2
                        debug_print("has dot and index %d" % dotidx, "to_abcmusic")
                        i += 1
                    else:
                        debug_print((i, f[i], f[i+1]), "flags")
                        for j in xrange(int(f[i+1])):
                            flags.append([1,int(f[i])])

                        i += 2

                debug_print((flags), "flags")

                hasbeams = True
            elif g.match_id_name("*flag*"):
                # get the value of the flag and calculate
                # its number for the abctab-output
                if flagcourse < 0:
                    flagcourse = g.course
                f=g.id_name[0][1].split('.')
                if f[-1] != "longa":
                    speed=int(f[-1])
                else:
                    speed = 1
                flags=[[1,speed]]
                flag_course=g.course


            elif g.match_id_name("*fret*"):
                while course < g.course:
                    course+=1
                f=g.id_name[0][1].split('.')
                fretletter=f[len(f)-1]
                chord.append(transcribe(\
                    fretletter, course, tuninglist))
                course+=1
                flag_course=-1

            elif g.match_id_name("*dot*"):
                debug_print("DOT", "to_abctab")
                if len(flags)>0 and\
                        g.gridnr <= len(flags) and g.course == flagcourse:
                        #                    g.gridnr > 0 and
                    #                        and g.course == flag_course:
                    if g.gridnr > 0: #dotgridnr
                        debug_print(( "B", flags), "to_abctab")
                        flags[g.gridnr-1][0] *= 3
                        flags[g.gridnr-1][1] *= 2
                    else:
                        debug_print(( "A", flags), "to_abctab")
                        flags[g.gridnr][0] *= 3
                        flags[g.gridnr][1] *= 2
                else:
                    #debug_print(("C", i, g.x_number, g.gridnr, g.course, flag_course, flags), "to_abctab")
                    dot+="."

        # don't forget the last chord
        if len(chord) > 0:
            #if last_bar != 0:
            #    song+="\\\n\t"
            #    last_bar=0

#            chord, no=self.uild_chord()
            #n_items+=no
            #n_chords+=1
            song+=self.build_chord(chord, flags, div)

        song+="\n%"
        return song




#----------------------------------------------------------------------------
    def to_abcmusic_german(self):
        """Generates abcmusic output for a germanglyph page
"""

        def transcribe(fret, course, tuning):

            if course > len(tuning):
                return ""

            # compute note and octave
            octave = tuning[course-1].octave
            note = self.scalenum[tuning[course-1].note]
            octave += (note + fret) / 12
            note = (note + fret) % 12

            musicnote = self.scalenote[note]
            #print "note '" + musicnote + "'",
            musicnote = self.accidentalize(musicnote,setaccs)
            #print "-> '" + musicnote + "'"

            if octave <= 0:
                musicnote = musicnote.upper()
                for i in range(octave,0):
                    musicnote += ","
            elif octave > 1:
                for i in range(1,octave):
                    musicnote += "'"

            return musicnote

        def getflagpart(flag,gr_number):
            # get the rhythm value for a note_group
            if flag==None or "fermata" in flag.get_main_id():
                return None
            #welchen wert hat das flag an der stelle gr_number
            name=flag.get_main_id().split(".")[1:]

            if len(name)%2==1:
                name.append(1)
            i=0
            s=[]
            # erzeuge liste der rhythmischen werte  und nehme die an der stelle die ich brauche
            while i<len(name):
                s.extend([int(name[i])]*int(name[i+1]))
                i+=2
            if gr_number>=len(s):
                return None
            if hasattr(flag,"dot") and gr_number==0:
                return [3,2*s[gr_number]]
            else:
                return [1,s[gr_number]]


        # guess the key such that minimal accidentals will result
        def guesskey(glyphs,tuning):

            # build up list of all notes (without octave info)
            notes = []
            for line_group in glyphs:
                for bar_group in line_group:
                    for flag_group in bar_group:
                        flag = flag_group[0][0]
                        if "flag" in flag.get_main_id():
                            note_groups = flag_group[1:]
                        else:
                            flag=None
                            note_groups = flag_group
                        for note_group in note_groups:
                            for note in note_group:
                                if "fret" in note.get_main_id():
                                    course=int(note.get_main_id()[6])
                                    fret=int(note.get_main_id()[9])
                                    noteind = (self.scalenum[tuning[course-1].note] + fret) % 12
                                    notes.append(self.scalenote[noteind])


            #test b's or sharp's
            return self.test_b_sharp(notes)

        # parse tuning into list of tuningcourses
        tuninglist=self.get_tuninglist()


        # guess key if wished
        if self.key == 'auto':
            self.key = guesskey(self.glyphs,tuninglist)
            print "key guessed as '" + self.key + "'"


        chord=[]
        flag=None
        old_flag=None
        div=2
        Lfieldval = 1.0/div * (2**self.factor)
        setaccs = []
        # write abc header
        song = self.song_header
        if Lfieldval >= 1:
            song += "L:" + str(int(Lfieldval)) + "/1\n"
            song += "Q:" + str(int(Lfieldval)) + "/1=120\n"
        else:
            song += "L:1/" + str(int(1.0/Lfieldval)) + "\n"
            song += "Q:1/" + str(int(1.0/Lfieldval)) + "=100\n"
        song += "K:" + self.key + "\n"+\
            "%\n"


        flag_course=-1
        #write abc_notes
        for line_group in self.glyphs:

            for bar_number,bar_group in enumerate(line_group):
                for flag_group in bar_group:
                    if flag!=None:
                        old_flag=flag
                    flag=flag_group[0][0]
                    if "flag" in flag.get_main_id():
                        note_groups = flag_group[1:]
                    else:
                        flag=None
                        note_groups = flag_group
                    for note_group_ind,note_group in enumerate(note_groups):
                        chord=[]
                        for note in note_group:
                            if "fret" in note.get_main_id():
                                course=int(note.get_main_id()[6])
                                fret=int(note.get_main_id()[9])
                                chord.append(transcribe(fret, course, tuninglist))
                        if len(chord) > 0:
                            flagpart=getflagpart(flag,note_group_ind)
                            if flagpart==None:
                                flagpart=getflagpart(old_flag,note_group_ind)
                            if flagpart!=None:
                                song+=self.build_chord(chord, [flagpart], div)

                if bar_group!=[]:
                    song+="|"
                    if bar_number!=len(line_group)-1:
                        song+="\\\n\t"

            if len(line_group)>0:
                song+="\n%\n"

        return song
#-------------------------------------------------------------------------------
